/**
 * Created by Evalds on 24/05/2017.
 */

const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const he = require('he'); // html entity decoder https://github.com/mathiasbynens/he

const UsersController = require('../controllers/UsersController');
const InstrumentsController = require('../controllers/InstrumentsController');
const ProjectController = require('../controllers/ProjectController');
const ProjectGroupsController = require('../controllers/ProjectGroupsController');
const ReportController = require('../controllers/ReportController');
const DateTimeUtils = require('../common/DateTimeUtils');
const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const PermissionUtils = require('../utils/PermissionUtils');
const i18n = require('../utils/i18n');

const EventTypes = require('../common/EventTypes');
const DevelopmentTypes = require('../common/DevelopmentTypes');
const RequestMessageTypes = require('../common/RequestMessageTypes');

const _ = require('lodash');

module.exports = class AuditTrailController {

    static get EventTypes() {
        return EventTypes;
    }

    static get DevelopmentTypes() {
        return DevelopmentTypes;
    }

    static get Default() {
        return {
            id: 0,
            event_type_id: 0,
            instrument_id: 0,
            project_id: 0,
            report_id: 0,
            project_group_id: 0,
            dev_type_id: 0,
            project_file_revision: false,
            report_file_revision: false,
            message_content: '',
            user_id: 0,
            created_time: false,
            created_date: false,

            report_hash: null, // Link by hash reports before report_id available
            project_hash: null,

            user: {}
        }
    }

    static getTableName(){
        return 'audit_trail';
    }


    static async create(auditTrail, sessionToken)
    {
        try
        {
            auditTrail.user_id = sessionToken.user ? sessionToken.user.id : 0;
            CodeUtils.overrideProperties(auditTrail, this.Default);
            CodeUtils.fillFieldsCreatedModified(auditTrail, sessionToken);
            auditTrail.message_content = he.decode(auditTrail.message_content);
            auditTrail = await CodeUtils.insertDB('audit_trail', auditTrail);
        }
        catch (exc)
        {
            logger.error('AuditTrailController.create');
            logger.error(exc);
        }

        return auditTrail;
    }

    static async getAll(query, offset, limit, sort, sessionToken)
    {
        let results = [];

        let sql = '';
        try
        {
            if(!query) {
                query = '';
            }

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            let order_by_options = {
                default_sort: 'date',
                lower_column: ['users', 'message_content'],
                change_columns: [{console_name: 'user', server_name: 'user$username'}]
            };

            let order_by = '';
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort, order_by_options);
            }

            let permissionFilter = PermissionUtils.getAllowedIds(sessionToken, 'projectGroups', 'projects.project_group_id', {allow_zero_ids: true, allow_null_ids: sessionToken.joint_can_manage_users});

            sql = SQL`SELECT ` + CodeUtils.getColumnNames('audit_trail', AuditTrailController.Default) +
                SQL`, count(*) OVER() AS total, ` +
                CodeUtils.getColumnNames('users', UsersController.Default, 'user$') + ', ' +
                CodeUtils.getColumnNames('instruments', InstrumentsController.Default, 'instrument$') + ', ' +
                CodeUtils.getColumnNames('projects', ProjectController.Default, 'project$') + ', ' +
                CodeUtils.getColumnNames('project_groups', ProjectGroupsController.Default, 'projectGroup$') + ', ' +
                CodeUtils.getColumnNames('reports', ReportController.Default, 'report$') +
                SQL` FROM audit_trail
                    LEFT JOIN users ON users.id = audit_trail.user_id
                    LEFT JOIN instruments ON instruments.id = audit_trail.instrument_id
                    LEFT JOIN projects ON projects.id = audit_trail.project_id
                    LEFT JOIN project_groups ON project_groups.id = audit_trail.project_group_id
                    LEFT JOIN reports ON reports.id = audit_trail.report_id
                    WHERE ` +
                    (queryInt ? SQL`audit_trail.id = ${queryInt}` : SQL`audit_trail.message_content ILIKE ${query} `) +
                permissionFilter +
                order_by +
                ((offset !== undefined) ? SQL` OFFSET ${offset}` : '') +
                ((limit !== undefined) ? SQL` LIMIT ${limit}` : '');
            let rows = await pgAsync.rows(sql);

            results = CodeUtils.getJoinData(rows);
        }
        catch (exc)
        {
            logger.error('AuditTrailController.getAll', sql);
            logger.error(exc);
        }

        return results;
    }

    static async getLastProjectRevFile(item_id, fileType){
        let results = false;
        let sql = '';
        try
        {
            let rev_type = '';
            let item_type = '';
            switch(fileType){
                case 'report':
                    rev_type = 'report_file_revision';
                    item_type = 'report_id';
                    break;
                case 'project':
                    rev_type = 'project_file_revision';
                    item_type = 'project_id';
                    break;
            }
            sql = SQL`SELECT * FROM audit_trail WHERE ` + item_type + SQL` = ${item_id} AND ` + rev_type + SQL` = true ORDER BY id DESC`
            results = await pgAsync.rows(sql);
        }
        catch (exc)
        {
            logger.error('AuditTrailController.getLastProjectRevFile', sql);
            logger.error(exc);
        }
        return results[0] ? results[0] : false;
    }

    static getArrayForExport(items, devTypeName, eventTypeName){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('timestamp'),
            i18n.localize('type'),
            i18n.localize('program'),
            i18n.localize('user'),
            i18n.localize('message'),
            i18n.localize('instrument'),
            i18n.localize('project_file_revision'),
            i18n.localize('project_name'),
            i18n.localize('report_file_revision'),
            i18n.localize('report_name'),
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.created_formated,
                eventTypeName[item.event_type_id],
                devTypeName[item.dev_type_id],
                item.user.username,
                item.message_content,
                (item.instrument.name === undefined) ? '' : item.instrument.name,
                (item.project_file_revision === true) ? i18n.localize('project_file_on_server') : '',
                (item.project.name === undefined) ? '' : item.project.name,
                (item.report_file_revision === true) ? i18n.localize('report_file_on_server') : '',
                (item.report.name === undefined) ? '' : item.report.name
            ]);
        }
        return data;
    }

    static async getById(item_id, sessionToken){
        let item = false;
        try
        {
            item = await CodeUtils.getById(this.getTableName(), item_id, {is_deleted: true});
            if(item !== false){
                item.user = (item.user_id == '0') ? {} : await UsersController.getById(item.user_id, {is_deleted: true});
                item.user = (item.user === false) ? {} : item.user;

                item.projectGroup = (item.project_group_id == '0') ? {} : await ProjectController.getById(item.project_group_id, sessionToken, {is_deleted: true});
                item.projectGroup = (item.projectGroup === false) ? {} : item.projectGroup;

                item.instrument = (item.instrument_id == '0') ? {} : await InstrumentsController.getById(item.instrument_id, sessionToken, {is_deleted: true});
                item.instrument = (item.instrument === false) ? {} : item.instrument;

                item.project = (item.project_id == '0') ? {} : await ProjectController.getById(item.project_id, sessionToken, {is_deleted: true});
                item.project = (item.project === false) ? {} : item.project;

                item.report = (item.report_id == '0') ? {} : await ReportController.getById(item.report_id, sessionToken, {is_deleted: true});
                item.report = (item.report === false) ? {} : item.report;
            }
        }
        catch (exc)
        {
            logger.error('AuditTrailController.getById');
            logger.error(exc);
        }

        return item;
    }

    static async getByAssoictedId(type, item_id, offset, limit, sessionToken){
        let results = [];

        let sql = '';
        try{
            let assoc_field_name = '';
            if(type == 'project'){
                assoc_field_name = this.getTableName()+'.project_id';
                let project = ProjectController.getById(item_id, sessionToken);
                if(project === false){
                    return false;
                }
            }
            else{
                assoc_field_name = this.getTableName()+'.report_id';
                let report = ReportController.getById(item_id, sessionToken);
                if(report === false){
                    return false;
                }
            }

            sql = SQL`SELECT ` + CodeUtils.getColumnNames('audit_trail', AuditTrailController.Default) +
                                            SQL`, count(*) OVER() AS total, ` +
                                            CodeUtils.getColumnNames('users', UsersController.Default, 'user$') + ', ' +
                                            CodeUtils.getColumnNames('instruments', InstrumentsController.Default, 'instrument$') + ', ' +
                                            CodeUtils.getColumnNames('projects', ProjectController.Default, 'project$') + ', ' +
                                            CodeUtils.getColumnNames('project_groups', ProjectGroupsController.Default, 'projectGroup$') + ', ' +
                                            CodeUtils.getColumnNames('reports', ReportController.Default, 'report$') +
                                            SQL` FROM audit_trail
                                                LEFT JOIN users ON users.id = audit_trail.user_id
                                                LEFT JOIN instruments ON instruments.id = audit_trail.instrument_id
                                                LEFT JOIN projects ON projects.id = audit_trail.project_id
                                                LEFT JOIN project_groups ON project_groups.id = audit_trail.project_group_id
                                                LEFT JOIN reports ON reports.id = audit_trail.report_id
                                                WHERE ` + assoc_field_name + SQL` = ${item_id} ` + SQL`ORDER BY audit_trail.id DESC` +
                                                ((offset !== undefined) ? SQL` OFFSET ${offset}` : '') +
                                                ((limit !== undefined) ? SQL` LIMIT ${limit}` : '');

            let rows = await pgAsync.rows(sql);

            results = CodeUtils.getJoinData(rows);
        }
        catch(exc)
        {
            logger.error('AuditTrailController.getByAssoictedId', sql);
            logger.error(exc);
        }

        return results;
    }
}
