/**
 * Created by Evalds on 24/05/2017.
 */

const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const InstrumentIpAddressesController = require('../controllers/InstrumentIpAddressesController');
const PermissionUtils = require('../utils/PermissionUtils');
const i18n = require('../utils/i18n');
const VendorTypes = require('../common/VendorTypes');

const _ = require('lodash');

module.exports = class InstrumentsController {

    static get Default() {
        return {
            id: 0,
            name: '',
            description: '',
            vendor: 0,

            is_deleted: false,
            modified_user_id: false,
            modified_time: false,
            modified_date: false,
            created_user_id: false,
            created_time: false,
            created_date: false
        }
    }


    static getTableName(){
        return 'instruments';
    }

    static async create(instrument, sessionToken)
    {
        try
        {
            let ip_addresses = instrument.ip_addresses;
            CodeUtils.overrideProperties(instrument, this.Default);
            CodeUtils.fillFieldsCreatedModified(instrument, sessionToken);

            instrument = await CodeUtils.insertDB(this.getTableName(), instrument);

            instrument.ip_addresses = [];

            if(instrument.id !== undefined && instrument.id > 0){
                instrument.ip_addresses = await InstrumentIpAddressesController.create(instrument.id, ip_addresses, sessionToken);
            }
            else{
                throw new ReferenceError("empty instrument.id in InstrumentsController.create");
            }
        }
        catch (exc)
        {
            logger.error('InstrumentsController.create');
            logger.error(exc);
        }

        return instrument;
    }


    static async getByPcNames(pc_names, vendor)
    {
        let instruments_ids = [];
        let sql = '';
        try
        {
            let vendor_id = 0;
            let sql_vendor = '';
            if(vendor) {
                if(vendor === 'WatersICS') {
                    vendor = 'Waters';
                }
                else if(vendor === 'ThermoChromeleon') {
                    vendor = 'Thermo';
                }
                else if(vendor === 'Hitachi') {
                    vendor = 'Hitachi';
                }
                else if(vendor === 'OpenLab' || vendor === 'OpenLabWithICF') {
                    vendor = 'OpenLab';
                }

                for(let key in VendorTypes) {
                    if(key === vendor) {
                        vendor_id = VendorTypes[key];
                        sql_vendor =
                            `INNER JOIN instruments 
                            ON instrument_ip_addresses.instrument_id = instruments.id AND vendor = ${vendor_id} `;
                        break
                    }
                }
            }

            sql = `SELECT instrument_ip_addresses.instrument_id FROM instrument_ip_addresses ${sql_vendor} 
                WHERE instrument_ip_addresses.host = ANY($1) GROUP BY instrument_ip_addresses.instrument_id`;

            let rows = await pgAsync.rows(sql, pc_names);
            if(rows.length === 0) {
                instruments_ids = false;
            }
            else{
                for (let row of rows){
                    instruments_ids.push(row.instrument_id);
                }
            }
        }
        catch (exc)
        {
            logger.error('InstrumentsController.getByPcName', sql);
            logger.error(exc);
        }

        return instruments_ids;
    }

    static async getAll(query, offset, limit, sort, sessionToken)
    {
        let instruments = [];
        let sql = '';
        try
        {
            if(!query) {
                query = '';
            }

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            query = '%' + query + '%';

            let order_by = '';
            let sql_limit = '';
            let sql_offset = '';
            if(limit !== undefined){
                sql_limit = `LIMIT ${limit}`;
            }
            if(offset !== undefined){
                sql_offset = `OFFSET ${offset}`;
            }
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort);
            }

            let permissionFilter = PermissionUtils.getAllowedIds(sessionToken, 'instruments', this.getTableName()+'.id');

            let sub_sql_modified_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.modified_user_id) as modified_user';
            let sub_sql_created_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.created_user_id) as created_user';
            let sub_sql_ips = '(SELECT string_agg(host, \', \' ORDER BY LOWER(host)) FROM instrument_ip_addresses ips WHERE ips.instrument_id = '+this.getTableName()+'.id AND is_deleted=false) AS pc_names_string';

            sql = `SELECT *, count(*) OVER() AS total, 
                ${sub_sql_ips}, 
                ${sub_sql_created_user}, 
                ${sub_sql_modified_user} FROM ${this.getTableName()} 
                WHERE 
                ${queryInt ? ' id = ' + queryInt.toString() + ' AND ' : ''}
                name ILIKE $1 AND is_deleted=FALSE ${permissionFilter} ${order_by} ${sql_limit} ${sql_offset}`;

            instruments = await pgAsync.rows(sql, query);

            let is_array = true;
            instruments = await CodeUtils.populateOne2Many('instrument_ip_addresses', 'instrument_id', 'ip_addresses', instruments, is_array);
        }
        catch (exc)
        {
            logger.error('InstrumentsController.getAll', sql);
            logger.error(exc);
        }

        return instruments;
    }

    // Each specific project must be accessible also after it has been deleted, only in lists it should not be visible (is_deleted flag ignored)
    static async getById(id, sessionToken, options){

        let _options = _.defaults(options || {}, {
            is_deleted: true
        });

        let instrument = false;

        try
        {
            instrument = await CodeUtils.getById(this.getTableName(), id, {is_deleted: _options.is_deleted});
            if(instrument !== false){
                instrument = await CodeUtils.populateOne2Many('instrument_ip_addresses', 'instrument_id', 'ip_addresses', [instrument]);
                instrument = PermissionUtils.checkAssocPermission(instrument, 'instruments', sessionToken);
            }
        }
        catch (exc)
        {
            logger.error('InstrumentsController.getById');
            logger.error(exc);
        }

        return instrument;
    }

    static async checkForUniqueName(name){
        try
        {
            var is_uniq = await CodeUtils.isUniqFieldData(this.getTableName(), 'name', name);
        }
        catch (exc)
        {
            logger.error('InstrumentsController.checkForUniqueRole');
            logger.error(exc);
        }

        return is_uniq;
    }

    static async update(instrument, sessionToken){
        try{
            let ip_addresses = instrument.ip_addresses;

            CodeUtils.overrideProperties(instrument, this.Default);
            CodeUtils.fillFieldsCreatedModified(instrument, sessionToken);

            instrument = await CodeUtils.updateDB(this.getTableName(), instrument);

            await InstrumentIpAddressesController.update(instrument.id, ip_addresses, sessionToken);
        }
        catch (exc)
        {
            logger.error('InstrumentsController.update');
            logger.error(exc);
        }

        return instrument;
    }

    static async delete(item_id, sessionToken){
        try
        {
            await CodeUtils.delete('id', item_id, sessionToken.user, this.getTableName());
            await InstrumentIpAddressesController.delete(item_id, sessionToken);

        }
        catch (exc)
        {
            logger.error('InstrumentsController.delete');
            logger.error(exc);
        }

        return true;
    }

    static async checkForDuplicatedProps(instrument){
        let items = await this.getAll('');
        let duplicate_vendor_count = 0;
        let duplicate_host_count = 0;
        for(let item of items){
            if(item.id == instrument.id){
                continue;
            }
            if(item.vendor == instrument.vendor){
                for(let instr_ips of instrument.ip_addresses){
                    for(let ips of item.ip_addresses){
                        if(instr_ips.host == ips.host){
                            duplicate_host_count++;
                            break;
                        }
                    }
                }
                if(item.ip_addresses.length == duplicate_host_count && item.ip_addresses.length == instrument.ip_addresses.length){
                    duplicate_vendor_count++;
                }
            }
        }
        if(duplicate_vendor_count > 0){
            return true;
        }
        return false;
    }

    static getArrayForExport(items, vendorTypeName){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('name'),
            i18n.localize('description'),
            i18n.localize('pc_name'),
            i18n.localize('vendor'),
            i18n.localize('modified_user'),
            i18n.localize('modified_date'),
            i18n.localize('modified_time'),
            i18n.localize('created_user'),
            i18n.localize('created_date'),
            i18n.localize('created_time')
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.name,
                item.description,
                item.pc_names_string,
                vendorTypeName[item.vendor],
                item.modified_user,
                item.modified_date,
                item.modified_time,
                item.created_user,
                item.created_date,
                item.created_time,
            ]);
        }
        return data;
    }
}
