/**
 * Created by Evalds on 13/06/2017.
 */


const UsersController = require('../controllers/UsersController');
const ProjectGroupsController = require('../controllers/ProjectGroupsController');
const InstrumentsController = require('../controllers/InstrumentsController');
const PermissionUtils = require('../utils/PermissionUtils');

const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const i18n = require('../utils/i18n');
const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const DevelopmentTypes = require('../common/DevelopmentTypes');

const _ = require('lodash');

module.exports = class ProjectController {

    static get Default() {
        return {
            id: 0,
            project_group_id: 0,
            dev_type_id: 0,
            instrument_id: 0,

            name: '',
            description: '',

            count_runs: 0,
            count_sequences: 0,

            is_deleted: false,
            modified_user_id: false,
            modified_time: false,
            modified_date: false,
            created_user_id: false,
            created_time: false,
            created_date: false
        }
    }

    static getTableName(){
        return 'projects';
    }

    static async isUniqueNameInProjectGroup(project)
    {
        let result = false;
        let sql = '';
        try
        {
            sql = SQL`SELECT * FROM projects 
                WHERE name = ${project.name} AND is_deleted = false AND 
                project_group_id = ${project.project_group_id} AND NOT id = ${project.id} LIMIT 1`;

            let data = await pgAsync.rows(sql);
            result = (data.length === 0);
        }
        catch (exc)
        {
            logger.error('ProjectController._avoidFileNameCollision');
            logger.error(exc);
        }

        return result;
    }

    static async create(project, sessionToken)
    {
        try
        {
            project.id = 0;
            CodeUtils.overrideProperties(project, ProjectController.Default);
            CodeUtils.fillFieldsCreatedModified(project, sessionToken);

            project = await CodeUtils.insertDB('projects', project);
        }
        catch (exc)
        {
            logger.error('ProjectController.create');
            logger.error(exc);
        }

        return project;
    }

    static async update(project, sessionToken)
    {
        try
        {

            CodeUtils.overrideProperties(project, ProjectController.Default);
            CodeUtils.fillFieldsCreatedModified(project, sessionToken);

            project = await CodeUtils.updateDB('projects', project);
        }
        catch (exc)
        {
            logger.error('ProjectController.update');
            logger.error(exc);
        }

        return project;
    }

    static async isProjectGroupUsing(project_group_id){
        try
        {
            var is_uniq = await CodeUtils.isUniqFieldData(this.getTableName(), 'project_group_id', project_group_id);
        }
        catch (exc)
        {
            logger.error('ProjectController.isProjectGroupUsing');
            logger.error(exc);
        }

        return is_uniq;
    }

    static async getAll(query, offset, limit, sort, sessionToken)
    {
        let items = [];
        let sql = '';
        try
        {
            if(!query) {
                query = '';
            }

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            query = '%' + query + '%';

            let order_by = '';
            let sql_limit = '';
            let sql_offset = '';
            if(limit !== undefined){
                sql_limit = `LIMIT ${limit}`;
            }
            if(offset !== undefined){
                sql_offset = `OFFSET ${offset}`;
            }
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort);
            }

            let permissionFilter = PermissionUtils.getAllowedIds(sessionToken, 'projectGroups', this.getTableName()+'.project_group_id');

            let sub_sql_projectGroups = '(SELECT name FROM project_groups pg WHERE pg.id = '+this.getTableName()+'.project_group_id) as project_group_name';
            let sub_sql_instrument = "(SELECT name FROM instruments i WHERE i.id = "+this.getTableName()+'.instrument_id) as instrument_name';
            let sub_sql_modified_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.modified_user_id) as modified_user';
            let sub_sql_created_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.created_user_id) as created_user';

            sql = `SELECT *, count(*) OVER() AS total, 
                ${sub_sql_projectGroups}, 
                ${sub_sql_instrument}, 
                ${sub_sql_modified_user}, 
                ${sub_sql_created_user} FROM ${this.getTableName()} 
                WHERE 
                ${queryInt ? ' id = ' + queryInt.toString() + ' AND ' : ''}
                name ILIKE $1 AND is_deleted=false ${permissionFilter} ${order_by} ${sql_limit} ${sql_offset}`;

            items = await pgAsync.rows(sql, query);

            let propName = function (prop, value){
               for(let i in prop) {
                   if (prop[i] === value){
                        return i;
                   }
               }
               return false;
            };

            for(let item of items){
                item.projectGroup = await ProjectGroupsController.getById(item.project_group_id, sessionToken);
                item.instrument = await InstrumentsController.getById(item.instrument_id, sessionToken);
                item.dev_type_name = propName(DevelopmentTypes, item.dev_type_id);
            }
        }
        catch (exc)
        {
            logger.error('ProjectController.getAll', sql);
            logger.error(exc);
        }

        return items;
    }

    // Each specific project must be accessible also after it has been deleted, only in lists it should not be visible (is_deleted flag ignored)
    static async getById(item_id, sessionToken, options){

        let _options = _.defaults(options || {}, {
            is_deleted: true
        });

        let item = undefined;
        try
        {
            item = await CodeUtils.getById(this.getTableName(), item_id, {is_deleted: _options.is_deleted});
            if(item !== false){
                item.projectGroup = await ProjectGroupsController.getById(item.project_group_id, sessionToken, {is_deleted: _options.is_deleted});
                item.instrument = await InstrumentsController.getById(item.instrument_id, sessionToken, {is_deleted: _options.is_deleted});
                item = PermissionUtils.checkAssocPermission(item, 'projects', sessionToken);
            }
        }
        catch (exc)
        {
            logger.error('ProjectController.getById');
            logger.error(exc);
        }

        return item;
    }

    static async delete(item_id, user){
        try
        {
            await CodeUtils.delete('id', item_id, user, this.getTableName());
        }
        catch (exc)
        {
            logger.error('ProjectController.delete');
            logger.error(exc);
        }

        return true;
    }

    static async isInstrumentUsing(id){
        try
        {
            var is_uniq = await CodeUtils.isUniqFieldData(this.getTableName(), 'instrument_id', id);
        }
        catch (exc)
        {
            logger.error('ProjectController.isInstrumentUsing');
            logger.error(exc);
        }

        return is_uniq;
    }

    static getArrayForExport(items, devTypeName){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('name'),
            i18n.localize('description'),
            i18n.localize('project_group'),
            i18n.localize('instrument'),
            i18n.localize('program'),
            i18n.localize('count_sequences'),
            i18n.localize('count_runs'),
            i18n.localize('modified_user'),
            i18n.localize('modified_date'),
            i18n.localize('modified_time'),
            i18n.localize('created_user'),
            i18n.localize('created_date'),
            i18n.localize('created_time')
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.name,
                item.description,
                item.project_group_name,
                item.instrument_name,
                devTypeName[item.dev_type_id],
                item.count_sequences,
                item.count_runs,
                item.modified_user,
                item.modified_date,
                item.modified_time,
                item.created_user,
                item.created_date,
                item.created_time,
            ]);
        }
        return data;
    }
}
