/**
 * Created by Evalds on 24/05/2017.
 */

const UsersController = require('../controllers/UsersController');
const PermissionUtils = require('../utils/PermissionUtils');
const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const i18n = require('../utils/i18n');

const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;

const _ = require('lodash');

module.exports = class ProjectGroupsController {

    static get Default() {
        return {
            id: 0,
            name: '',
            description: '',

            is_deleted: false,
            modified_user_id: false,
            modified_time: false,
            modified_date: false,
            created_user_id: false,
            created_time: false,
            created_date: false
        }
    }

    static getTableName(){
        return 'project_groups';
    }

    static async create(projectGroups, sessionToken){
        try
        {
            CodeUtils.overrideProperties(projectGroups, this.Default);
            CodeUtils.fillFieldsCreatedModified(projectGroups, sessionToken);

            projectGroups = await CodeUtils.insertDB(this.getTableName(), projectGroups);

        }
        catch (exc)
        {
            logger.error('ProjectGroupsController.create');
            logger.error(exc);
        }

        return projectGroups;
    }

    static async update(project, sessionToken)
    {
        try
        {
            CodeUtils.overrideProperties(project, ProjectGroupsController.Default);
            CodeUtils.fillFieldsCreatedModified(project, sessionToken);

            project = await CodeUtils.updateDB(this.getTableName(), project);
        }
        catch (exc)
        {
            logger.error('ProjectController.create');
            logger.error(exc);
        }

        return project;
    }

    static async getAvailableProjects(sessionToken)
    {
        let results = [];
        let sql = '';

        try
        {
            sql = SQL`select * from project_groups WHERE is_deleted=false ORDER by name ASC`;
            let project_groups = await pgAsync.rows(sql);

            for(let project_group of project_groups) {
                let isAdded = false;
                if (sessionToken.user.can_access_all_project_groups) {
                    results.push(project_group);
                    continue;
                }
                else {
                    for (let projectGroup of sessionToken.user.projectGroups) {
                        if (projectGroup.id === project_group.id) {
                            results.push(project_group);
                            isAdded = true;
                            break;
                        }
                    }
                    if(isAdded) {
                        continue;
                    }
                }

                for (let u_r of sessionToken.user.roles) {
                    if (u_r.can_access_all_project_groups) {
                        results.push(project_group);
                        isAdded = true;
                        break;
                    }

                    for (let projectGroup of u_r.projectGroups) {
                        if (projectGroup.id === project_group.id) {
                            results.push(project_group);
                            isAdded = true;
                            break;
                        }
                    }
                    if(isAdded) {
                        break;
                    }
                }
            }
        }
        catch (exc)
        {
            logger.error('ProjectGroupsController.getAvailableProjects', sql);
            logger.error(exc);
        }

        return results;
    }

    static async getAll(query, offset, limit, sort, sessionToken)
    {
        let items = [];
        let sql = '';

        try
        {
            if(!query) {
                query = '';
            }
            query = '%' + query + '%';

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            let order_by = '';
            let sql_limit = '';
            let sql_offset = '';
            if(limit !== undefined){
                sql_limit = `LIMIT ${limit}`;
            }
            if(offset !== undefined){
                sql_offset = `OFFSET ${offset}`;
            }
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort);
            }

            let permissionFilter = PermissionUtils.getAllowedIds(sessionToken, 'projectGroups', this.getTableName()+'.id');

            let sub_sql_modified_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.modified_user_id) as modified_user';
            let sub_sql_created_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.created_user_id) as created_user';

            sql = `SELECT *, count(*) OVER() AS total, 
                ${sub_sql_created_user}, 
                ${sub_sql_modified_user} FROM ${this.getTableName()} 
                WHERE
                 ${queryInt ? ' id = ' + queryInt.toString() + ' AND ' : ''}
                name ILIKE $1 AND is_deleted=false ${permissionFilter} ${order_by} ${sql_limit} ${sql_offset}`;

            items = await pgAsync.rows(sql, query);

        }
        catch (exc)
        {
            logger.error('ProjectGroupsController.getAll', sql);
            logger.error(exc);
        }

        return items;
    }

    // Each specific project must be accessible also after it has been deleted, only in lists it should not be visible (is_deleted flag ignored)
    static async getById(item_id, sessionToken, options){
        let _options = _.defaults(options || {}, {
            is_deleted: true
        });

        let item = false;

        try
        {
            item = await CodeUtils.getById(this.getTableName(), item_id, {is_deleted: _options.is_deleted});
            if(item !== false){
                item = PermissionUtils.checkAssocPermission(item, 'projectGroups', sessionToken);
            }
        }
        catch (exc)
        {
            logger.error('ProjectGroupsController.getById');
            logger.error(exc);
        }

        return item;
    }

    static async checkForUniqueName(name){
        try
        {
            var is_uniq = await CodeUtils.isUniqFieldData(this.getTableName(), 'name', name);
        }
        catch (exc)
        {
            logger.error('ProjectGroupsController.checkForUniqueName');
            logger.error(exc);
        }

        return is_uniq;
    }

    static async delete(item_id, user){
        try
        {
            await CodeUtils.delete('id', item_id, user, this.getTableName());
        }
        catch (exc)
        {
            logger.error('ProjectGroupsController.delete');
            logger.error(exc);
            return false;
        }

        return true;
    }

    static getArrayForExport(items, devTypeName){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('name'),
            i18n.localize('description'),
            i18n.localize('modified_user'),
            i18n.localize('modified_date'),
            i18n.localize('modified_time'),
            i18n.localize('created_user'),
            i18n.localize('created_date'),
            i18n.localize('created_time')
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.name,
                item.description,
                item.modified_user,
                item.modified_date,
                item.modified_time,
                item.created_user,
                item.created_date,
                item.created_time,
            ]);
        }
        return data;
    }
}
