const ProjectController = require('../controllers/ProjectController');
const InstrumentsController = require('../controllers/InstrumentsController');
const PermissionUtils = require('../utils/PermissionUtils');

const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const i18n = require('../utils/i18n');
const logger = require('../common/Logger');
const DateTimeUtils = require('../common/DateTimeUtils');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const DevelopmentTypes = require('../common/DevelopmentTypes');
const EventTypes = require('../common/EventTypes');

const _ = require('lodash');

module.exports = class ReportController {

    static get Default() {
        return {
            id: 0,
            project_id: 0,
            dev_type_id: 0,

            name: '',
            description: '',

            hash: '', //Unique identificator to sync id, given by client, taken from specially marked session token

            count_runs: 0,
            count_sequences: 0,
            count_run_minutes: 0,

            is_finished: false,
            is_failed: false,

            is_deleted: false,
            modified_user_id: false,
            modified_time: false,
            modified_date: false,
            created_user_id: false,
            created_time: false,
            created_date: false,

            project: {}
        }
    }

    static async updateAuditTrailByOfflineHash(report)
    {
        let sql = '';
        try {
            if(report.hash && report.project) {
                if (report.hash.length > 0) {
                    sql = SQL`UPDATE audit_trail SET 
                            report_id = ${report.id},
                            instrument_id = ${report.instrument_id},
                            project_id = ${report.project.id},
                            project_group_id = ${report.project.project_group_id} 
                            WHERE report_id = 0 AND report_hash = ${report.hash}`;
                    let result = await pgAsync.query(sql);
                }
            }
        }
        catch (exc) {
            logger.error('ReportController.updateAuditTrailByOfflineHash', sql);
            logger.error(exc);
        }
    }

    static getTableName() {
        return 'reports';
    }

    static async isUniqueNameInReports(item)
    {
        let result = false;
        let sql = '';
        try
        {
            sql = `SELECT * FROM ${this.getTableName()} 
                WHERE name = $1 AND 
                is_deleted = false AND 
                project_id = $2 AND NOT id = $3 LIMIT 1`;

            let data = await pgAsync.rows(
                sql,
                item.name,
                item.project_id,
                item.id
            );
            result = (data.length === 0);
        }
        catch (exc)
        {
            logger.error('ReportController._avoidFileNameCollision', sql);
            logger.error(exc);
        }

        return result;
    }

    // hash as an alternative identification for reports when submitted in async manner after session expiration
    static async isExistingHash(hash)
    {
        let result = false;
        let sql = '';
        try
        {
            sql = `SELECT * FROM reports WHERE hash = $1 LIMIT 1;`;
            let items = await pgAsync.rows(sql, hash);

            if(items.length) {
                result = true;
            }
            else {
                logger.warn('ReportController.isExistingHash missing offline report by hash: ' + hash);
            }
        }
        catch (exc)
        {
            logger.error('ReportController.isExistingHash', sql);
            logger.error(exc);
        }

        return result;
    }

    // hash as an alternative identification for reports when submitted in async manner after session expiration
    static async getByHash(hash, sessionToken)
    {
        let report = undefined;
        let sql = '';
        try
        {
            sql = `SELECT * FROM reports WHERE hash = $1 LIMIT 1;`;
            let items = await pgAsync.rows(sql, hash);

            if(items.length) {
                report = items[0];
                let project = await ProjectController.getById(report.project_id, sessionToken);
                if(project) {
                    report.project = project;
                    report.project_group_id = report.project.project_group_id;
                    report = PermissionUtils.checkAssocPermission(report, 'reports', sessionToken);
                }
                else {
                    logger.error('ReportController.getByHash missing project_id: ' + report.project_id.toString());
                }
            }
            else {
                logger.error('ReportController.getByHash missing offline report by hash: ' + hash);
            }
        }
        catch (exc)
        {
            logger.error('ReportController.getByHash', sql);
            logger.error(exc);
        }

        return report;
    }

    static async create(report, sessionToken)
    {
        try
        {
            report.id = 0;
            CodeUtils.overrideProperties(report, ReportController.Default);
            CodeUtils.fillFieldsCreatedModified(report, sessionToken);

            report = await CodeUtils.insertDB(this.getTableName(), report);
            report.project = await ProjectController.getById(report.project_id, sessionToken);
        }
        catch (exc)
        {
            logger.error('ReportController.create');
            logger.error(exc);
        }

        return report;
    }

    static async update(item, sessionToken)
    {
        try
        {
            if(item.project === false) {
                item.project = undefined;
            }
            item = CodeUtils.setOnlyDefaultProperties(item, ReportController.Default);
            CodeUtils.fillFieldsCreatedModified(item, sessionToken);

            await CodeUtils.updateDB(this.getTableName(), item);
            item = await ReportController.getById(item.id, sessionToken);
            item.project = await ProjectController.getById(item.project_id, sessionToken);
        }
        catch (exc)
        {
            logger.error('ReportController.update');
            logger.error(exc);
        }

        return item;
    }

    static async getAll(query, offset, limit, sort, sessionToken)
    {
        let items = [];
        let sql = '';
        try
        {
            if(!query) {
                query = '';
            }

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            query = '%' + query + '%';

            let order_by = '';
            let sql_limit = '';
            let sql_offset = '';
            if(limit !== undefined){
                sql_limit = `LIMIT ${limit}`;
            }
            if(offset !== undefined){
                sql_offset = `OFFSET ${offset}`;
            }
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort, {table_name: this.getTableName()});
            }
            let sub_sql_project = '(SELECT projects.name FROM projects p WHERE p.id = '+this.getTableName()+'.project_id) as project_name';

            let permissionFilter = PermissionUtils.getAllowedIds(sessionToken, 'projectGroups', 'projects.project_group_id');

            let sub_sql_modified_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.modified_user_id) as modified_user';
            let sub_sql_created_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.created_user_id) as created_user';

            sql = `SELECT reports.*, count(*) OVER() AS total, 
                ${this.getTableName()}.id AS _id, 
                ${sub_sql_project}, 
                ${sub_sql_modified_user}, 
                ${sub_sql_created_user},
                instruments.name as instrument_name
                FROM ${this.getTableName()} 
                JOIN projects ON projects.id = ${this.getTableName()}.project_id 
                JOIN instruments ON instruments.id = ${this.getTableName()}.instrument_id
                WHERE 
                 ${queryInt ? ' reports.id = ' + queryInt.toString() + ' AND ' : ''}
                reports.name ILIKE $1 and reports.is_deleted=false 
                ${permissionFilter} ${order_by} ${sql_limit} ${sql_offset}`;

            items = await pgAsync.rows(sql, query);

            let propName = function (prop, value){
               for(let i in prop) {
                   if (prop[i] === value){
                        return i;
                   }
               }
               return false;
            };

            for(let item of items){
                item.project = await ProjectController.getById(item.project_id, sessionToken);
                item.project_group_id = item.project.project_group_id;
                item.dev_type_name = propName(DevelopmentTypes, item.dev_type_id);
            }
        }
        catch (exc)
        {
            logger.error('ReportController.getAll', sql);
            logger.error(exc);
        }

        return items;
    }

    static async getById(item_id, sessionToken, options){
        let _options = _.defaults(options || {}, {
            is_deleted: true
        });

        let item = undefined;

        try
        {
            item = await CodeUtils.getById(this.getTableName(), item_id, {is_deleted: _options.is_deleted});
            if(item !== false){
                item.project = (item.instrument_id === 0) ? {} : await ProjectController.getById(item.project_id, sessionToken, {is_deleted: _options.is_deleted});
                item.project = (item.project === false) ? {} : item.project;

                item.instrument = (item.instrument_id === 0) ? {} : await InstrumentsController.getById(item.instrument_id, sessionToken, {is_deleted: _options.is_deleted});
                item.instrument = (item.instrument === false) ? {} : item.instrument;

                item = PermissionUtils.checkAssocPermission(item, 'reports', sessionToken);
            }
        }
        catch (exc)
        {
            logger.error('ReportController.getById');
            logger.error(exc);
        }

        return item;

    }

    static async delete(item_id, user){
        try
        {
            await CodeUtils.delete('id', item_id, user, this.getTableName());
        }
        catch (exc)
        {
            logger.error('ReportController.delete');
            logger.error(exc);
        }

        return true;
    }

    static getArrayForExport(items, devTypeName){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('name'),
            i18n.localize('description'),
            i18n.localize('program'),
            i18n.localize('project'),
            i18n.localize('count_sequences'),
            i18n.localize('count_runs'),
            i18n.localize('is_failed'),
            i18n.localize('is_finished'),
            i18n.localize('modified_user'),
            i18n.localize('modified_date'),
            i18n.localize('modified_time'),
            i18n.localize('created_user'),
            i18n.localize('created_date'),
            i18n.localize('created_time')
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.name,
                item.description,
                item.project_name,
                devTypeName[item.dev_type_id],
                item.count_sequences,
                item.count_runs,
                item.is_failed,
                (item.is_finished === undefined) ? '-' : item.is_finished,
                item.modified_user,
                item.modified_date,
                item.modified_time,
                item.created_user,
                item.created_date,
                item.created_time,
            ]);
        }
        return data;
    }

    static async getProgramUsageOverview(selectedProjectGroups, dateFrom, dateTo, sessionToken){
        let items = [];
		let sql = '';
        try
        {
            let pg_ids = [];
            for(let j of selectedProjectGroups){
                pg_ids.push(j.id);
            }
			
			if(pg_ids.length)
			{
				let sql_case = 'CASE '+this.getTableName()+'.dev_type_id';
				for(let i of Object.keys(DevelopmentTypes)){
					sql_case += ' WHEN ' + DevelopmentTypes[i] + ' THEN \'' + i18n.localize(i) +'\'' ;
				}
				sql_case += ' END as program';

                let sql_dataRange = '';
                if(dateFrom.length){
                    sql_dataRange += " AND "+this.getTableName()+".created_date >= '"+dateFrom.slice(0,10)+"'";
                }
                if(dateTo.length){
                    sql_dataRange +=" AND "+this.getTableName()+".created_date <= '"+dateTo.slice(0,10)+"'";
                }

				sql = SQL`SELECT ` + sql_case + SQL`, sum(reports.count_runs) as runs` +
											SQL` FROM ` + this.getTableName() +
											SQL` INNER JOIN projects ON `+this.getTableName()+SQL`.project_id = projects.id` +
											SQL` WHERE projects.project_group_id IN (` + pg_ids.join() + SQL`) AND `+this.getTableName()+SQL`.is_deleted=false` + sql_dataRange +
											SQL` GROUP BY `+this.getTableName()+SQL`.dev_type_id`;

				items = await pgAsync.rows(sql);
			}
        }
        catch (exc)
        {
            logger.error('ReportController.getProgramUsageOverview', sql);
			logger.error(sql);
            logger.error(exc);
        }

        return items;
    }

    static async getActiveUsersOverview(selectedProjectGroups, dateFrom, dateTo, sessionToken){
        let items = [];
        let sql = '';
        try
        {
            let pg_ids = [];
            for(let j of selectedProjectGroups){
                pg_ids.push(j.id);
            }

			if(pg_ids.length)
			{
				let sql_case = "(CASE WHEN ((users.firstname IS NULL OR users.firstname = '') AND (users.surname IS NULL OR users.surname = '')) THEN users.username ELSE users.firstname || ' ' || users.surname END) as full_name";

                let sql_dataRange = '';
                if(dateFrom.length){
                    sql_dataRange += " AND "+this.getTableName()+".created_date >= '"+dateFrom.slice(0,10)+"'";
                }
                if(dateTo.length){
                    sql_dataRange +=" AND "+this.getTableName()+".created_date <= '"+dateTo.slice(0,10)+"'";
                }

				sql = SQL`SELECT `+sql_case+`, sum(reports.count_runs) as runs` +
											SQL` FROM ` + this.getTableName() +
											SQL` INNER JOIN projects ON `+this.getTableName()+SQL`.project_id = projects.id` +
											SQL` INNER JOIN users ON `+this.getTableName()+SQL`.created_user_id = users.id` +
											SQL` WHERE projects.project_group_id IN (` + pg_ids.join() + SQL`) AND `+this.getTableName()+SQL`.is_deleted=false` + sql_dataRange +
											SQL` GROUP BY users.id`;

				items = await pgAsync.rows(sql);
			}
        }
        catch (exc)
        {
            logger.error('ReportController.getActiveUsersOverview', sql);
            logger.error(exc);
        }

        return items;
    }

    static async getInstrumentsUsageOverview(selectedProjectGroups, dateFrom, dateTo, sessionToken){
        let items = [];
        let sql = '';

        try
        {
            let pg_ids = [];
            for(let j of selectedProjectGroups){
                pg_ids.push(j.id);
            }

			if(pg_ids.length)
			{

                let sql_dataRange = ' AND '+this.getTableName()+'.created_date IS NOT NULL ';
				if(dateFrom.length){
                    sql_dataRange += " AND "+this.getTableName()+".created_date >= '"+dateFrom.slice(0,10)+"'";
				}
                if(dateTo.length){
                    sql_dataRange +=" AND "+this.getTableName()+".created_date <= '"+dateTo.slice(0,10)+"'";
                }

				sql =   SQL`SELECT DISTINCT ON (reports.instrument_id) reports.instrument_id, instruments.name` +
							SQL` FROM reports ` +
							SQL` INNER JOIN instruments ON reports.instrument_id = instruments.id ` +
                            SQL` INNER JOIN projects ON projects.id = reports.project_id ` +
							SQL` WHERE projects.project_group_id IN (` + pg_ids.join() + SQL`) AND reports.is_deleted=false`;
				let sub_items = await pgAsync.rows(sql);

				let values = [];

				for(let item of sub_items){
					sql =   SQL`SELECT reports.created_date as date, sum(reports.count_run_minutes) as minutes` +
							SQL` FROM reports` +
							SQL` WHERE instrument_id = `+item.instrument_id+SQL` AND is_deleted = false` + sql_dataRange +
							SQL` GROUP BY reports.created_date`;
					values = await pgAsync.rows(sql);

					if(values.length > 0){
					    let itemFormated = {
                            key: item.name,
                            values: []
                        };
					    for(let each of values) {
					        if(each.minutes > 0) {
					            each.hours =  _.round(each.minutes / 60, 1); //to hours
                                itemFormated.values.push(each);
                            }
                        }
                        if(itemFormated.values.length > 0) {
                            items.push(itemFormated);
                        }
					}
				}
			}
        }
        catch (exc)
        {
            logger.error('ReportController.getInstrumentsUsageOverview', sql);
            logger.error(exc);
        }

        return items;
    }


    static async getRunsOverview(selectedProjectGroups, dateFrom, dateTo, sessionToken){
        let items = [];
        let sql = '';

        try
        {
            let pg_ids = [];
            for(let j of selectedProjectGroups){
                pg_ids.push(j.id);
            }

			if(pg_ids.length)
			{
				let sql_case = 'CASE audit_trail.event_type_id';
				for(let i of Object.keys(EventTypes)){
					if(EventTypes[i] == 0 || EventTypes[i] == 1){
						sql_case += ' WHEN ' + EventTypes[i] + ' THEN \'' + i18n.localize('success') +'\'' ;
					}
					else{
						sql_case += ' WHEN ' + EventTypes[i] + ' THEN \'' + i18n.localize(i) +'\'' ;
					}

				}
				sql_case += ' END as status';

                let sql_dataRange = '';
                if(dateFrom.length){
                    sql_dataRange += " AND "+this.getTableName()+".created_date >= '"+dateFrom.slice(0,10)+"'";
                }
                if(dateTo.length){
                    sql_dataRange +=" AND "+this.getTableName()+".created_date <= '"+dateTo.slice(0,10)+"'";
                }

				sql = SQL`SELECT `+sql_case+SQL`, sum(reports.count_runs) as runs` +
												SQL` FROM ` + this.getTableName() +
												SQL` INNER JOIN projects ON `+this.getTableName()+SQL`.project_id = projects.id` +
												SQL` INNER JOIN audit_trail ON `+this.getTableName()+SQL`.id = audit_trail.report_id` +
												SQL` WHERE projects.project_group_id IN (` + pg_ids.join() + SQL`) AND `+this.getTableName()+SQL`.is_deleted=false` + sql_dataRange +
												SQL` GROUP BY audit_trail.event_type_id`;

				items = await pgAsync.rows(sql);
			}
        }
        catch (exc)
        {
            logger.error('ReportController.getRunsOverview', sql);
            logger.error(exc);
        }

        return items;
    }


    static async getProjectGroupsOverview(selectedProjectGroups, dateFrom, dateTo, sessionToken){
        let items = [];
        let sql = '';

        try
        {
            let pg_ids = [];
            for(let j of selectedProjectGroups){
                pg_ids.push(j.id);
            }

			if(pg_ids.length)
			{
				let sql_case = "(CASE WHEN (instruments.name IS NULL OR instruments.name = '') THEN instruments.name ELSE instruments.name END) as instrument_name";

                let sql_dataRange = '';
                if(dateFrom.length){
                    sql_dataRange += " AND "+this.getTableName()+".created_date >= '"+dateFrom.slice(0,10)+"'";
                }
                if(dateTo.length){
                    sql_dataRange +=" AND "+this.getTableName()+".created_date <= '"+dateTo.slice(0,10)+"'";
                }

				sql =   SQL`SELECT project_groups.id as id, project_groups.name as name` +
							SQL` FROM project_groups` +
							SQL` WHERE project_groups.id IN (` + pg_ids.join() + SQL`) AND project_groups.is_deleted=false`;
				let sub_items = await pgAsync.rows(sql);

				let values = [];

				for(let item of sub_items){
					sql =   SQL`SELECT reports.created_date as x, sum(reports.count_runs) as y` +
							SQL` FROM reports` +
							SQL` INNER JOIN projects ON reports.project_id = projects.id` +
							SQL` WHERE projects.project_group_id = ${item.id} AND reports.is_deleted = false` + sql_dataRange +
							SQL` GROUP BY reports.created_date`;
					values = await pgAsync.rows(sql);
					if(values.length > 0){
						items.push({
							key: item.name,
							values: values
						});
					}
				}
			}
        }
        catch (exc)
        {
            logger.error('ReportController.getProjectGroupsOverview', sql);
            logger.error(exc);
        }

        return items;
    }
}
