/**
 * Created by Evalds on 24/05/2017.
 */

const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const RolesInstrumentsController = require('../controllers/RolesInstrumentsController');
const RolesProjectGroupsController = require('../controllers/RolesProjectGroupsController');
const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const i18n = require('../utils/i18n');
const _ = require('lodash');
const DateTimeUtils = require("../common/DateTimeUtils");

module.exports = class RolesController {

    static get Default() {
        return {
            id: 0,
            name: '',

            is_deleted: false,
            modified_user_id: false,
            modified_time: false,
            modified_date: false,
            created_user_id: false,
            created_time: false,
            created_date: false,
            can_access_admin_console: false,
            can_manage_users: false,
            can_access_all_project_groups: false,
            can_access_all_instruments: false,
        }
    }

    static getTableName(){
        return 'roles';
    }

    static async create(role, sessionToken)
    {
        try
        {
            var currentDate = (new Date());
            var currentTime = DateTimeUtils.getTimeLocalString(currentDate);
            let result = await pgAsync.query(
                SQL`INSERT INTO roles
                    (
                        name,
                        is_deleted,
                        modified_user_id,
                        modified_time,
                        modified_date,
                        created_user_id,
                        created_time,
                        created_date,
                        can_access_admin_console,
                        can_manage_users,
                        can_access_all_project_groups,
                        can_access_all_instruments
                    )
                    VALUES
                    (
                        ${role.name},
                        false,
                        ${sessionToken.user.id},
                        ${currentTime},
                        ${currentDate},
                        ${sessionToken.user.id},
                        ${currentTime},
                        ${currentDate},
                        ${role.can_access_admin_console},
                        ${role.can_manage_users},
                        ${role.can_access_all_project_groups},
                        ${role.can_access_all_instruments}
                    ) RETURNING id`);

            role.id = result.rows[0].id;

            role.roles_instruments = [];
            role.roles_project_groups = [];

            role.roles_instruments = await RolesInstrumentsController.create(role.id, role.instruments, sessionToken.user);
            role.roles_project_groups = await RolesProjectGroupsController.create(role.id, role.projectGroups, sessionToken.user);
        }
        catch (exc)
        {
            logger.error('RolesController.create');
            logger.error(exc);
        }

        return role;
    }

    static async checkForUniqueRole(role_name){
        try
        {
            var is_uniq = await CodeUtils.isUniqFieldData(this.getTableName(), 'name', role_name);
        }
        catch (exc)
        {
            logger.error('RolesController.checkForUniqueRole');
            logger.error(exc);
        }

        return is_uniq;
    }

    static async getAll(query, offset, limit, sort)
    {
        let roles = [];
        let sql = '';

        try
        {
            if(!query) {
                query = '';
            }

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            query = '%' + query + '%';

            let order_by = '';
            let sql_limit = '';
            let sql_offset = '';
            if(limit !== undefined){
                sql_limit = `LIMIT ${limit}`;
            }
            if(offset !== undefined){
                sql_offset = `OFFSET ${offset}`;
            }
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort);
            }

            let sub_sql_modified_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.modified_user_id) as modified_user';
            let sub_sql_created_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.created_user_id) as created_user';
            let sub_sql_instruments = '(SELECT string_agg(name, \', \' ORDER BY LOWER(name)) FROM instruments i JOIN roles_instruments ON (i.id = roles_instruments.instrument_id) WHERE roles_instruments.role_id = '+this.getTableName()+'.id AND roles_instruments.is_deleted=false) AS concat_instrument_names';
            let sub_sql_projectGroups = '(SELECT string_agg(name, \', \' ORDER BY LOWER(name)) FROM project_groups i JOIN roles_project_groups ON (i.id = roles_project_groups.project_group_id) WHERE roles_project_groups.role_id = '+this.getTableName()+'.id AND roles_project_groups.is_deleted=false) AS concat_project_group_names';

            sql = `SELECT *, count(*) OVER() AS total, ${sub_sql_instruments}, ${sub_sql_projectGroups}, ${sub_sql_created_user}, ${sub_sql_modified_user} FROM ${this.getTableName()}             
                    WHERE 
                    ${queryInt ? ' id = ' + queryInt.toString() + ' AND ' : ''}
                    name ILIKE $1 AND is_deleted=FALSE ${order_by} ${sql_limit} ${sql_offset}`;
            roles = await pgAsync.rows(sql, query);

            for (let role of roles){
                role.instruments = await CodeUtils.getForMany2Many(role.id, 'roles_instruments', 'role_id', 'instruments', 'instrument_id');
                role.projectGroups = await CodeUtils.getForMany2Many(role.id, 'roles_project_groups', 'role_id', 'project_groups', 'project_group_id');
            }
        }
        catch (exc)
        {
            logger.error('RolesController.getAll', sql);
            logger.error(exc);
        }

        return roles;
    }

    static async getById(role_id, options){
        let role = false;

        let _options = _.defaults(options || {}, {
            is_details: true,
            is_deleted: true
        });

        role = await CodeUtils.getById(this.getTableName(), role_id, _options);
        if(_options.is_details && role !== false){
            role.instruments = await CodeUtils.getForMany2Many(role_id, 'roles_instruments', 'role_id', 'instruments', 'instrument_id');
            role.projectGroups = await CodeUtils.getForMany2Many(role_id, 'roles_project_groups', 'role_id', 'project_groups', 'project_group_id');
        }
        return role;
    }

    static async update(role, sessionToken){
        try
        {

            CodeUtils.overrideProperties(role, RolesController.Default);
            CodeUtils.fillFieldsCreatedModified(role, sessionToken);

            role = await CodeUtils.updateDB('roles', role);

            await RolesInstrumentsController.update(role.id, role.instruments, sessionToken.user);
            await RolesProjectGroupsController.update(role.id, role.projectGroups, sessionToken.user);
        }
        catch (exc)
        {
            logger.error('RolesController.update');
            logger.error(exc);
        }

        return role;
    }

    static async delete(role_id, user){
        try
        {
            await CodeUtils.delete('id', role_id, user, this.getTableName());
            await RolesInstrumentsController.delete(role_id, -1, user);
            await RolesProjectGroupsController.delete(role_id, -1, user);
        }
        catch (exc)
        {
            logger.error('RolesController.delete');
            logger.error(exc);
        }

        return true;
    }

    static async isInstrumentUsing(id){
        try
        {
            var is_uniq = RolesInstrumentsController.isInstrumentUsing(id);
        }
        catch (exc)
        {
            logger.error('RolesController.isInstrumentUsing');
            logger.error(exc);
        }

        return is_uniq;
    }

    static getArrayForExport(items){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('name'),
            i18n.localize('can_manage_users'),
            i18n.localize('can_access_admin_console'),
            i18n.localize('instruments'),
            i18n.localize('project_groups'),
            i18n.localize('modified_user'),
            i18n.localize('modified_date'),
            i18n.localize('modified_time'),
            i18n.localize('created_user'),
            i18n.localize('created_date'),
            i18n.localize('created_time')
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.name,
                (item.can_manage_users == true) ? i18n.localize('yes') : i18n.localize('no'),
                (item.can_access_admin_console == true) ? i18n.localize('yes') : i18n.localize('no'),
                (item.can_access_all_instruments == true) ? i18n.localize('all') : item.concat_instrument_names,
                (item.can_access_all_project_groups == true) ? i18n.localize('all') : item.concat_project_group_names,
                item.modified_user,
                item.modified_date,
                item.modified_time,
                item.created_user,
                item.created_date,
                item.created_time,
            ]);
        }
        return data;
    }
}
