const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const DateTimeUtils = require('../common/DateTimeUtils');
const CodeUtils = require('../utils/CodeUtils');
const PermissionUtils = require('../utils/PermissionUtils');

const md5 = require('crypto-js/md5');
const encHex = require('crypto-js/enc-hex');

const UsersController = require('../controllers/UsersController');

// Only requests with valid session token can be executed
// As requested from FDA part 11 it must be valid only for 15min of inactivity
// All audit trail and data storing must continue after this timeout

module.exports = class SessionTokensController
{
    static get Default()
    {
        return {
            id : 0,
            is_valid: false,
            hash: '',
            user_id: 0,
            dev_type_id: 0,
            created_date: (new Date()),
            created_time: DateTimeUtils.getTimeLocalString(new Date()),
            last_used_date: new Date(),
            last_used_time: DateTimeUtils.getTimeLocalString(new Date()),

            user: UsersController.Default,
            last_used: (new Date()),
            created: (new Date()),

            offline_project_id: 0, // used for report data gathering after session has been terminated
            offline_instrument_id: 0
        };
    }

    // Create new session
    static async create(user, request)
    {
        let sessionToken = this.Default;

        try
        {
            sessionToken.user_id = user.id;
            sessionToken.is_valid = true;
            sessionToken.hash = md5('cs' + (new Date()).toISOString()).toString(encHex);
            if(!request.dev_type_id) {
                request.dev_type_id = 0;
            }

            let result = await pgAsync.query(
                SQL`INSERT INTO session_tokens
                    (
                        is_valid,
                        hash,
                        user_id,
                        created_date,
                        created_time,
                        last_used_date,
                        last_used_time,
                        dev_type_id
                    )
                    VALUES
                    (
                        ${sessionToken.is_valid},
                        ${sessionToken.hash},
                        ${sessionToken.user_id},
                        ${sessionToken.created_date},
                        ${sessionToken.created_time},
                        ${sessionToken.last_used_date},
                        ${sessionToken.last_used_time},
                        ${request.dev_type_id}
                    ) RETURNING id`);

            sessionToken.last_used = DateTimeUtils.combine(sessionToken.last_used_date, sessionToken.last_used_time);
            sessionToken.created = DateTimeUtils.combine(sessionToken.created_date, sessionToken.created_time);

            sessionToken.id = result.rows[0].id;
            sessionToken.user = user;
            sessionToken = PermissionUtils.joinAccessProps(sessionToken);
        }
        catch (exc)
        {
            logger.error('SessionTokensController.create');
            logger.error(exc);
        }

        return sessionToken;
    }


    static async update(sessionToken)
    {
        try
        {
            CodeUtils.overrideProperties(sessionToken, this.Default);
            sessionToken.last_used_date = this.Default.last_used_date;
            sessionToken.last_used_time = this.Default.last_used_time;

            await pgAsync.query(
                SQL`UPDATE session_tokens SET
                        last_used_date = ${sessionToken.last_used_date},
                        last_used_time = ${sessionToken.last_used_time},
                        offline_project_id = ${sessionToken.offline_project_id},
                        offline_instrument_id = ${sessionToken.offline_instrument_id},
                        is_valid = ${sessionToken.is_valid}
                    WHERE id = ${sessionToken.id}`);

            sessionToken = PermissionUtils.joinAccessProps(sessionToken);
        }
        catch (exc)
        {
            logger.error('SessionTokensController.update');
            logger.error(exc);
        }
    }


    static async getById(id)
    {
        let sessionToken = this.Default;

        try
        {
            sessionToken = await pgAsync.row(SQL`select * from session_tokens where id=${id} limit 1;`);

            sessionToken.last_used = DateTimeUtils.combine(sessionToken.last_used_date, sessionToken.last_used_time);
            sessionToken.created = DateTimeUtils.combine(sessionToken.created_date, sessionToken.created_time);
            sessionToken.user = await UsersController.getById(sessionToken.user_id);
            sessionToken = PermissionUtils.joinAccessProps(sessionToken);

        }
        catch (exc)
        {
            logger.error('SessionTokensController.getById');
            logger.error(exc);
        }

        return sessionToken;
    }

    static async getByHash(hash)
    {
        let sessionToken = this.Default;

        try
        {
            sessionToken = await pgAsync.row(SQL`select * from session_tokens where hash=${hash} limit 1;`);

            sessionToken.last_used = DateTimeUtils.combine(sessionToken.last_used_date, sessionToken.last_used_time);
            sessionToken.created = DateTimeUtils.combine(sessionToken.created_date, sessionToken.created_time);
            sessionToken.user = await UsersController.getById(sessionToken.user_id, {is_details: true});
            sessionToken = PermissionUtils.joinAccessProps(sessionToken);

        }
        catch (exc)
        {
            logger.error('SessionTokensController.getById');
            logger.error(exc);
        }

        return sessionToken;
    }
}
