/**
 * Created by Evalds on 24/05/2017.
 */

const logger = require('../common/Logger');
const pgAsync = require('../utils/Database');
const SQL = require('pg-async').SQL;
const crypto = require('crypto');
const CodeUtils = require('../utils/CodeUtils');
const GeneralUtils = require('../utils/GeneralUtils');
const UsersInstrumentsController = require('../controllers/UsersInstrumentsController');
const UsersProjectGroupsController = require('../controllers/UsersProjectGroupsController');
const UsersRolesController = require('../controllers/UsersRolesController');
const i18n = require('../utils/i18n');
const _ = require('lodash');
const DateTimeUtils = require("../common/DateTimeUtils");

const USER_PASSWORD_SALT = 'swc';

module.exports = class UsersController {

    static get Default()
    {
        return {
            id: 0,
            username: '',
            password: '',
            firstname: '',
            surname: '',
            job_title: '',
            password_time: false,
            password_date: false,
            is_active: false,

            is_deleted: false,
            modified_user_id: false,
            modified_time: false,
            modified_date: false,
            created_user_id: false,
            created_time: false,
            created_date: false,
            can_access_admin_console: false,
            can_manage_users: false,
            can_access_all_project_groups: false,
            can_access_all_instruments: false
        }
    }

    static getTableName(){
        return 'users';
    }

    static async create(user, sessionToken)
    {
        try
        {
            var hash = crypto.createHmac('sha256', USER_PASSWORD_SALT)
                .update(user.password + user.username)
                .digest('hex');

            user.password = hash;
            user.is_active = true;
            CodeUtils.overrideProperties(user, this.Default);
            CodeUtils.fillFieldsCreatedModified(user, sessionToken);

            user = await CodeUtils.insertDB(this.getTableName(), user);

            user.users_instruments = [];
            user.users_project_groups = [];

            user.users_instruments = await UsersInstrumentsController.create(user.id, user.instruments, sessionToken.user);
            user.users_project_groups = await UsersProjectGroupsController.create(user.id, user.projectGroups, sessionToken.user);
            user.users_roles = await UsersRolesController.create(user.id, user.roles, sessionToken.user);

        }
        catch (exc)
        {
            logger.error('UsersController.create');
            logger.error(exc);
        }

        return user;
    }

    static async getByUsernamePassword(username, password)
    {
        let user = this.Default;

        try
        {
            var hash = crypto.createHmac('sha256', USER_PASSWORD_SALT)
                .update(password + username)
                .digest('hex');

            // For testing & getting hashes
            //console.log(hash);

            let results = await pgAsync.rows(SQL`select * from users where username=${username} and password=${hash} limit 1;`);
            if(results.length)
            {
                user = results[0];
                await this.populateData(user);
            }
        }
        catch (exc)
        {
            logger.error('UsersController.getByUsernamePassword');
            logger.error(exc);
        }

        return user;
    }

    static async getById(id, options)
    {
        let _options = _.defaults(options || {}, {
            is_details : true,
            is_deleted: true
        });
        let user = false;

        user = await CodeUtils.getById(this.getTableName(), id, {is_deleted: _options.is_deleted});
        if(_options.is_details){
            await this.populateData(user);
        }

        return user;
    }

    // Add all associated data to object
    static async populateData(user)
    {
        if(user !== false){
            user.instruments = await CodeUtils.getForMany2Many(user.id, 'users_instruments', 'user_id', 'instruments', 'instrument_id');
            user.projectGroups = await CodeUtils.getForMany2Many(user.id, 'users_project_groups', 'user_id', 'project_groups', 'project_group_id');
            user.roles = await CodeUtils.getForMany2Many(user.id, 'users_roles', 'user_id', 'roles', 'role_id');
            if(user.roles !== false){
                for(let role of user.roles){
                    role.projectGroups = await CodeUtils.getForMany2Many(role.id, 'roles_project_groups', 'role_id', 'project_groups', 'project_group_id');
                    role.instruments = await CodeUtils.getForMany2Many(role.id, 'roles_instruments', 'role_id', 'instruments', 'instrument_id');
                }
            }
        }
    }

    static async getAll(query, offset, limit, sort)
    {
        let users = [];
        let sql = '';

        try
        {
            if(!query)
            {
                query = '';
            }

            let queryInt = parseInt(query);
            if(!(queryInt > 0)) {
                query = '%' + query + '%';
                queryInt = 0;
            }
            else {
                query = '';
            }

            query = '%' + query + '%';

            let order_by = '';
            let sql_limit = '';
            let sql_offset = '';
            if(limit !== undefined){
                sql_limit = `LIMIT ${limit}`;
            }
            if(offset !== undefined){
                sql_offset = `OFFSET ${offset}`;
            }
            if(sort !== undefined){
                order_by = GeneralUtils.getOrderBy(sort);
            }

            let sub_sql_modified_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.modified_user_id) as modified_user';
            let sub_sql_created_user = '(SELECT username FROM users p WHERE p.id = '+this.getTableName()+'.created_user_id) as created_user';
            let sub_sql_instruments = '(SELECT string_agg(name, \', \' ORDER BY LOWER(name)) FROM instruments i JOIN users_instruments ON (i.id = users_instruments.instrument_id) WHERE users_instruments.user_id = users.id AND users_instruments.is_deleted=false) AS concat_instrument_names';
            let sub_sql_projectGroups = '(SELECT string_agg(name, \', \' ORDER BY LOWER(name)) FROM project_groups i JOIN users_project_groups ON (i.id = users_project_groups.project_group_id) WHERE users_project_groups.user_id = users.id AND users_project_groups.is_deleted=false) AS concat_project_group_names';
            let sub_sql_roles = '(SELECT string_agg(name, \', \' ORDER BY LOWER(name)) FROM roles i JOIN users_roles ON (i.id = users_roles.role_id) WHERE users_roles.user_id = users.id AND users_roles.is_deleted=false) AS concat_role_names';
            let sub_sql_fullname_case = "CASE WHEN (firstname IS NULL) OR (firstname = '') OR (surname is null) OR (surname = '') THEN LOWER(username) ELSE LOWER(firstname || ' ' || surname) END AS fullname";

            sql = `SELECT *, count(*) OVER() AS total, 
                ${sub_sql_fullname_case}, 
                ${sub_sql_instruments}, 
                ${sub_sql_projectGroups}, 
                ${sub_sql_roles}, 
                ${sub_sql_created_user}, 
                ${sub_sql_modified_user} 
                FROM ${this.getTableName()} WHERE 
                 ${queryInt ? ' id = ' + queryInt.toString() + ' AND ' : ''}
                username ILIKE $1 AND is_deleted=false ${order_by} ${sql_limit} ${sql_offset}`;

            users = await pgAsync.rows(sql, query);

            for (let user of users){
                await this.populateData(user);
            }
        }
        catch (exc)
        {
            logger.error('UsersController.getAll', sql);
            logger.error(exc);
        }

        return users;
    }

    static async checkForUniqueName(username){

        let is_uniq = false;
        try
        {
            is_uniq = await CodeUtils.isUniqFieldData(this.getTableName(), 'username', username);
        }
        catch (exc)
        {
            logger.error('UsersController.checkForUniqueName');
            logger.error(exc);
        }

        return is_uniq;
    }

    static async update(user, session_user){

        let sql = '';
        try
        {
            let currentDate = (new Date());
            let currentTime = DateTimeUtils.getTimeLocalString(currentDate);

            let user_columns_array = [
                'username',
                'firstname',
                'surname',
                'job_title',
                'can_access_admin_console',
                'can_manage_users',
                'can_access_all_project_groups',
                'can_access_all_instruments',
                'modified_date',
                'modified_time',
                'modified_user_id'
            ];

            let values = [
                user.username,
                user.firstname,
                user.surname,
                user.job_title,
                user.can_access_admin_console,
                user.can_manage_users,
                user.can_access_all_project_groups,
                user.can_access_all_instruments,
                currentDate,
                currentTime,
                session_user.id
            ];

            if(user.password !== ''){
                let hash = crypto.createHmac('sha256', USER_PASSWORD_SALT)
                    .update(user.password + user.username)
                    .digest('hex');
                user.password = hash;
                user.password_date = currentDate;
                user.password_time = currentTime;
                user_columns_array.push('password', 'password_date', 'password_time');
                values.push(
                    user.password,
                    user.password_date,
                    user.password_time
                );
            }

            let object = {
                'user': user,
                'currentDate': currentDate,
                'currentTime': currentTime,
                'session_user': session_user
            };

            sql = `UPDATE users 
                SET (${user_columns_array.join(', ')}) = 
                (${CodeUtils.generateSqlTemplatesFromValues(user_columns_array).join(', ')}) WHERE id = ${user.id};`;

            let result = await pgAsync.query(sql, ...values);

            await UsersInstrumentsController.update(user.id, user.instruments, session_user);
            await UsersProjectGroupsController.update(user.id, user.projectGroups, session_user);
            await UsersRolesController.update(user.id, user.roles, session_user);
        }
        catch (exc)
        {
            logger.error('UsersController.update', sql);
            logger.error(exc);
        }

        return user;
    }

    static async delete(user_id, session_user){
        try
        {
            await CodeUtils.delete('id', user_id, session_user, 'users');
            await UsersInstrumentsController.delete(user_id, -1, session_user);
            await UsersProjectGroupsController.delete(user_id, -1, session_user);
            await UsersRolesController.delete(user_id, -1, session_user);
        }
        catch (exc)
        {
            logger.error('UsersController.delete');
            logger.error(exc);
        }

        return true;
    }

    static async updatePassword(user, password){
        let sql = '';
        try
        {
            let currentDate = (new Date());
            let currentTime = DateTimeUtils.getTimeLocalString(currentDate);

            let hash = crypto.createHmac('sha256', USER_PASSWORD_SALT)
                .update(password + user.username)
                .digest('hex');

            let values = [
                hash,
                currentDate,
                currentTime
            ];

            sql = `UPDATE ${this.getTableName()} 
                SET (password, password_date, password_time) = 
                ($1, $2, $3) WHERE id = ${user.id};`;

            let result = await pgAsync.query(sql, ...values);

        }
        catch (exc)
        {
            logger.error('UsersController.updatePassword', sql);
            logger.error(exc);
            return false;
        }

        return true;
    }

    static getHash(username, password){
        return crypto.createHmac('sha256', USER_PASSWORD_SALT)
            .update(password + username)
            .digest('hex');
    }

    static async isInstrumentUsing(id){
        try
        {
            var is_uniq = UsersInstrumentsController.isInstrumentUsing(id);
        }
        catch (exc)
        {
            logger.error('UsersController.isInstrumentUsing');
            logger.error(exc);
        }

        return is_uniq;
    }

    static getArrayForExport(items){
        let data = [];
        data.push([
            i18n.localize('id'),
            i18n.localize('username'),
            i18n.localize('firstname'),
            i18n.localize('surname'),
            i18n.localize('job_title'),
            i18n.localize('password_date'),
            i18n.localize('password_time'),
            i18n.localize('can_manage_users'),
            i18n.localize('can_access_admin_console'),
            i18n.localize('instruments'),
            i18n.localize('project_groups'),
            i18n.localize('roles'),
            i18n.localize('modified_user'),
            i18n.localize('modified_date'),
            i18n.localize('modified_time'),
            i18n.localize('created_user'),
            i18n.localize('created_date'),
            i18n.localize('created_time')
        ]);
        for(let item of items){
            data.push([
                item.id,
                item.username,
                item.firstname,
                item.surname,
                item.job_title,
                item.password_date,
                item.password_time,
                (item.can_manage_users == true) ? i18n.localize('yes') : i18n.localize('no'),
                (item.can_access_admin_console == true) ? i18n.localize('yes') : i18n.localize('no'),
                (item.can_access_all_instruments == true) ? i18n.localize('all') : item.concat_instrument_names,
                (item.can_access_all_project_groups == true) ? i18n.localize('all') : item.concat_project_group_names,
                item.concat_role_names,
                item.modified_user,
                item.modified_date,
                item.modified_time,
                item.created_user,
                item.created_date,
                item.created_time,
            ]);
        }
        return data;
    }
}
