/**
 * Created by Evalds on 25/05/2017.
 */

const logger = require('../common/Logger');
const DateTimeUtils = require('../common/DateTimeUtils');
const SQL = require('pg-async').SQL;
const pgAsync = require('./Database');
const _ = require('lodash');

module.exports = class CodeUtils
{
    static generateSqlTemplatesFromValues(values) {
        let templates = [];
        let paramCount = 1;
        for(let value of values) {
            templates.push(`$${paramCount}`);
            paramCount ++;
        }
        return templates;
    }

    // Prepare data for inserting in DB
    static overrideProperties(original, defaults)
    {
        for (var column in defaults){
            if(original[column] === undefined)
            {
                original[column] = defaults[column];
            }
        }
    }

    // Prepare data for inserting in DB
    static setOnlyDefaultProperties(original, defaults)
    {
        for (var column in defaults){
            if(original[column] !== undefined)
            {
                defaults[column] = original[column];
            }
        }
        return defaults;
    }

    // For modified_user_id / created_user_id / modified_time / modified_date ..
    static fillFieldsCreatedModified(object, sessionToken)
    {
        let currentDate = (new Date());
        let currentTime = DateTimeUtils.getTimeLocalString(currentDate);

        //When offline token used time must be taken from request, because they come in with delay
        if(sessionToken.offline_timestamp !== undefined) {
            currentDate = sessionToken.offline_timestamp.date;
            currentTime = sessionToken.offline_timestamp.time;
        }

        if(object.dev_type_id !== undefined) {
            if(sessionToken.dev_type_id) {
                object.dev_type_id = sessionToken.dev_type_id;
            }
        }

        if(object.modified_user_id !== undefined) {
            object.modified_user_id = sessionToken.user.id;
        }

        if(object.modified_time !== undefined) {
            object.modified_time = currentTime;
        }

        if(object.modified_date !== undefined) {
            object.modified_date = currentDate;
        }

        if(object.password_date !== undefined&& !object.password_date) {
            object.password_date = currentDate;
        }

        if(object.password_time !== undefined && !object.password_time) {
            object.password_time = currentTime;
        }

        if(object.created_user_id !== undefined && !object.created_user_id) {
            object.created_user_id = sessionToken.user.id;
        }

        if(object.created_time !== undefined && !object.created_time) {
            object.created_time = currentTime;
        }

        if(object.created_date !== undefined && !object.created_date) {
            object.created_date = currentDate;
        }
    }

    // Insert object in DB
    static async insertDB(tableName, params, ignoreColumnsList)
    {
        let query = '';
        try {
            let columns = [];
            let templates = [];
            let values = [];
            let paramCounter = 1;


            for (let column in params) {

                if(ignoreColumnsList) {
                    if(_.includes(ignoreColumnsList, column)) {
                        continue;
                    }
                }

                if(tableName === 'reports') {
                    if(column === 'project_group_id') {
                        continue;
                    }
                }

                if (column !== 'id') {
                    //Only simple types
                    if (typeof params[column] !== 'object') {

                        columns.push(column);
                        values.push(params[column]);
                        templates.push('$' + paramCounter.toString());
                        paramCounter++;
                    }
                }
            }

            query = `INSERT INTO ${tableName} (${columns.join(',')}) VALUES (${templates.join(',')}) RETURNING id;`;
            let result = await pgAsync.query(query, ...values);

            params.id = result.rows[0].id;
        }
        catch (exc) {
            logger.error('CodeUtils.insertDB.query', query);
            logger.error('CodeUtils.insertDB.params', JSON.stringify(params));
            logger.error(exc);

            //Try to recover error
            let errorCol = CodeUtils.getErrorColumnFromExc(exc);
            if(errorCol) {
                if(!ignoreColumnsList || !_.includes(ignoreColumnsList, column)) {
                    logger.error('CodeUtils.insertDB.trying.recover.errorCol', errorCol);
                    if(!ignoreColumnsList) {
                        ignoreColumnsList = [];
                    }
                    ignoreColumnsList.push(errorCol);
                    return CodeUtils.insertDB(tableName, params, ignoreColumnsList);
                }
            }
        }

        return params;
    }

    // Update object in DB
    static async updateDB(tableName, params, ignoreColumnsList)
    {
        let query = '';
        try {

            let columns = [];
            let templates = [];
            let values = [];
            let paramCounter = 1;

            for (let column in params) {

                if(ignoreColumnsList) {
                    if(_.includes(ignoreColumnsList, column)) {
                        continue;
                    }
                }

                if (column !== 'id') {
                    //Only simple types
                    if (typeof params[column] !== 'object') {
                        columns.push(column);
                        values.push(params[column]);
                        templates.push('$' + paramCounter.toString());
                        paramCounter++;
                    }
                }
            }

            query = `UPDATE ${tableName} SET (${columns.join(',')}) = (${templates.join(',')}) WHERE id = $${paramCounter};`;

            values.push(params.id);
            let result = await pgAsync.query(query, ...values);
        }
        catch (exc) {
            logger.error('CodeUtils.updateDB.query', query);
            logger.error('CodeUtils.updateDB.params', JSON.stringify(params));
            logger.error(exc);

            //Try to recover error
            let errorCol = CodeUtils.getErrorColumnFromExc(exc);
            if(errorCol) {
                if(!ignoreColumnsList || !_.includes(ignoreColumnsList, column)) {
                    logger.error('CodeUtils.updateDB.trying.recover.errorCol', errorCol);
                    if(!ignoreColumnsList) {
                        ignoreColumnsList = [];
                    }
                    ignoreColumnsList.push(errorCol);
                    return CodeUtils.updateDB(tableName, params, ignoreColumnsList);
                }
            }
        }

        return params;
    }

    static getErrorColumnFromExc(exc) {
        let col = false;
        try {
            let strExc = exc.toString();
            let key = 'column "';
            if(strExc.indexOf(key) >= 0) {
                let idxStart = strExc.indexOf(key) + key.length;
                let idxEnd = strExc.indexOf('"', idxStart);
                col = strExc.substr(idxStart, idxEnd-idxStart);
            }
        }
        catch (exc) {
            logger.error('CodeUtils.getErrorColumnFromExc');
            logger.error(exc);
        }
        return col;
    }

    // Format column names
    static getColumnNames(table, templateObject, prefix = "")
    {
        var cols = [];

        for (var column in templateObject) {
            if(typeof templateObject[column] != 'object') {
                cols.push(table + '.' + column + ' AS ' + prefix + column);
            }
        }

        return cols.join(',');
    }

    // Joins prefixed with "table_name$"
    static getJoinData(rows)
    {
        var results = [];
        for(var row of rows) {
            var result = {};

            var parts = [];
            var partName = '';

            var currentObject = result;
            for (var column in row) {
                var keyName = column;

                if (column.indexOf('$') >= 0) {
                    partName = column.substring(0, column.indexOf('$'));
                    if(result[partName] === undefined) {
                        result[partName] = {};
                    }
                    currentObject = result[partName];
                    keyName = column.substring(column.indexOf('$') + 1);

                    if(parts.indexOf(currentObject) === -1) {
                        parts.push(currentObject);
                    }
                }

                if(keyName != 'password') {
                    currentObject[keyName] = row[column];
                }
            }

            parts.push(result);

            for(var each of parts) {
                if(each.created_date && each.created_time) {
                    each.created = DateTimeUtils.combine(each.created_date, each.created_time);
                    each.created_formated = each.created.toLocaleString();
                }
                if(each.modified_date && each.modified_time) {
                    each.modified = DateTimeUtils.combine(each.modified_date, each.modified_time);
                    each.modified_formated = each.modified.toLocaleString();
                }
                if(each.password_date && each.password_time) {
                    each.password_datetime = DateTimeUtils.combine(each.password_date, each.password_time);
                    each.password_datetime_formated = each.password_datetime.toLocaleString();
                }
            }

            results.push(result);
        }

        return results;
    }

    // id to avoid false when checking itself
    static async isUniqFieldData(table_name, field_name, field_data, id){
        let data = [];
        let sqlCondition = '';
        if(id) {
            sqlCondition = ' AND NOT id = ' + id.toString();
        }
        let query = `SELECT * FROM ${table_name} WHERE ${field_name} = $1 AND is_deleted = FALSE ${sqlCondition} LIMIT 1;`;
        data = await pgAsync.rows(query, field_data);
        return data.length === 0;
    }

    static async getForMany2Many(item_id, m2m_table_name, m2m_id_field_name, assoc_table_name, assoc_item_id_name){
        let object_ids = [];
        let result_object = [];
        let sql = '';

        try
        {
            sql = SQL`SELECT * FROM ` + m2m_table_name + SQL` WHERE ` + m2m_id_field_name + SQL` = ${item_id} AND is_deleted = FALSE;`;
            object_ids = await pgAsync.rows(sql);

            for(let object_id of object_ids){
                result_object.push(await this.getById(assoc_table_name, object_id[assoc_item_id_name]));
            }
        }
        catch (exc)
        {
            logger.error('CodeUtils.getForMany2Many', sql);
            logger.error(exc);
        }

        return result_object;
    }

    static async getById(table_name, item_id, options){
        let actual = _.defaults(options || {}, {
            is_deleted : false
        });

        let object = {};
        if(item_id <= 0 || item_id === undefined){
            return false;
        }

        let sql = '';
        try
        {
            sql = `SELECT * FROM ${table_name} WHERE id = $1
             ${(actual.is_deleted ? ' ' : ' AND is_deleted = false ')}
            LIMIT 1;`;
            object = await pgAsync.rows(sql, item_id);
        }
        catch (exc)
        {
            logger.error('CodeUtils.getById', sql);
            logger.error(exc);
        }

        if(object.length > 0){
            return object[0];
        }
        else {
            return false;
        }
    }

    static async delete(main_item_name, main_item_id, user, table_name, options){

        let _options = _.defaults(options || {}, {
            sub_item_name: false,
            sub_item_id: -1
        });

        let sql = '';
        try
        {
            let currentDate = (new Date());
            let currentTime = DateTimeUtils.getTimeLocalString(currentDate);
            let appendix = '';

            if(_options.sub_item_id >= 1){
                appendix = ` AND ${_options.sub_item_name} = ${_options.sub_item_id}`;
            }

            sql = `UPDATE ` + table_name + ` SET 
                (is_deleted, 
                modified_user_id, 
                modified_time, 
                modified_date) = 
                (true, 
                $1, 
                $2, 
                $3) WHERE ` +
                main_item_name + ` = ${main_item_id} ` + appendix;

            let result = await pgAsync.query(
                sql,
                user.id,
                currentTime,
                currentDate
            );
        }
        catch (exc)
        {
            logger.error('CodeUtils.delete', sql);
            logger.error(exc);
        }

        return true;
    }

    static async updateMany2Many(item_id, new_objects, user, existing_objects, del_func, create_func){
        let object = [];
        try
        {
            var currentDate = (new Date());
            var currentTime = DateTimeUtils.getTimeLocalString(currentDate);

            for (let new_object of new_objects)
            {
                new_object.is_presented = false;
                for (let existing_object of existing_objects)
                {
                    if(new_object.id == existing_object.id){
                        existing_object.is_presented = true;
                        new_object.is_presented = true;
                    }
                }
            }

            for (let existing_object of existing_objects)
            {
                if(existing_object.is_presented !== true){
                    await del_func(item_id, existing_object.id, user);
                }
            }

            for (let new_object of new_objects)
            {
                if(new_object.is_presented == false){
                    await create_func(item_id, [new_object], user);
                }
            }
        }
        catch (exc)
        {
            logger.error('CodeUtils.update');
            logger.error(exc);
        }

        return object;
    }

    static async createMany2Many(main_item_name, main_item_id, assoc_item_name, assoc_objects, table_name, user)
    {
        let many2many_object = [];
        let sql = '';
        try
        {
            let currentDate = (new Date());
            let currentTime = DateTimeUtils.getTimeLocalString(currentDate);
            for (let assoc_object of assoc_objects) {

                let values = [
                    currentTime,
                    currentDate,
                    user.id,
                    user.id,
                    currentTime,
                    currentDate,
                    main_item_id,
                    assoc_object.id
                ];
                sql = `INSERT INTO ${table_name}
                                (created_time, 
                                created_date, 
                                created_user_id, 
                                is_deleted, 
                                modified_user_id, 
                                modified_time, 
                                modified_date, 
                                ${main_item_name}, 
                                ${assoc_item_name})
                                VALUES (
                                $1, 
                                $2, 
                                $3, 
                                false, 
                                $4, 
                                $5, 
                                $6, 
                                $7, 
                                $8 ) RETURNING id;`;
                let result = await pgAsync.query(sql, ...values);

                many2many_object.push({
                    'created_time': currentTime,
                    'created_date': currentDate,
                    'created_user_id': user.id,
                    assoc_item_name: assoc_object.id,
                    'is_deleted': false,
                    'modified_user_id': user.id,
                    'modified_time': currentTime,
                    'modified_date': currentDate,
                    'role_id': main_item_id,
                    'id': result.rows[0].id
                });
            }

        }
        catch (exc)
        {
            logger.error('CodeUtils.createMany2Many', sql);
            logger.error(exc);
        }

        return many2many_object;
    }

    static async populateOne2Many(table_name, assoc_object_column_name, assoc_object_name, objects, is_array = false){

        let sql = '';
        try
        {
            for(let object of objects){
                if(!_.isEmpty(object)){
                    sql = `SELECT * FROM ` + table_name + ` WHERE 
                    ${assoc_object_column_name} = ${object.id} AND is_deleted = false;`;

                    object[assoc_object_name] = await pgAsync.rows(sql);
                }
            }
        }
        catch (exc)
        {
            logger.error('CodeUtils.populateOne2Many', sql);
            logger.error(exc);
        }

        if(objects.length === 1 && !is_array){
            return objects[0];
        }
        else{
            return objects;
        }
    }

    static async getOne2Many(table_name, assoc_object_column_name, assoc_object_id){
        let objects = [];

        let sql = '';
        try
        {
            sql = SQL`SELECT * FROM ` + table_name + SQL` WHERE 
            ${assoc_object_column_name} = ${assoc_object_id} AND is_deleted = false;`;
            objects = await pgAsync.rows(sql);

        }
        catch (exc)
        {
            logger.error('CodeUtils.populateOne2Many', sql);
            logger.error(exc);
        }
        return objects;
    }
}
